@@ -1,4 +1,3 @@
-#!/usr/bin/python3 -u
 #
 # dulwich - Simple command-line interface to Dulwich
 # Copyright (C) 2008-2011 Jelmer Vernooij <jelmer@jelmer.uk>
@@ -21,28 +20,29 @@
 # License, Version 2.0.
 #
 
-"""Simple command-line interface to Dulwich>
+"""Simple command-line interface to Dulwich>.
 
 This is a very simple command-line wrapper for Dulwich. It is by
 no means intended to be a full-blown Git command-line interface but just
 a way to test Dulwich.
 """
 
-import os
-import sys
-from getopt import getopt
 import argparse
 import optparse
+import os
 import signal
-from typing import Dict, Type, Optional
+import sys
+from getopt import getopt
+from typing import ClassVar, Dict, Optional, Type
 
 from dulwich import porcelain
-from dulwich.client import get_transport_and_path
-from dulwich.errors import ApplyDeltaError
-from dulwich.index import Index
-from dulwich.objectspec import parse_commit
-from dulwich.pack import Pack, sha_to_hex
-from dulwich.repo import Repo
+
+from .client import GitProtocolError, get_transport_and_path
+from .errors import ApplyDeltaError
+from .index import Index
+from .objectspec import parse_commit
+from .pack import Pack, sha_to_hex
+from .repo import Repo
 
 
 def signal_int(signal, frame):
@@ -55,7 +55,7 @@ def signal_quit(signal, frame):
     pdb.set_trace()
 
 
-class Command(object):
+class Command:
     """A Dulwich subcommand."""
 
     def run(self, args):
@@ -71,7 +71,7 @@ class cmd_archive(Command):
             type=str,
             help="Retrieve archive from specified remote repo",
         )
-        parser.add_argument('committish', type=str, nargs='?')
+        parser.add_argument("committish", type=str, nargs="?")
         args = parser.parse_args(args)
         if args.remote:
             client, path = get_transport_and_path(args.remote)
@@ -83,8 +83,7 @@ class cmd_archive(Command):
             )
         else:
             porcelain.archive(
-                ".", args.committish, outstream=sys.stdout.buffer,
-                errstream=sys.stderr
+                ".", args.committish, outstream=sys.stdout.buffer, errstream=sys.stderr
             )
 
 
@@ -107,8 +106,8 @@ class cmd_rm(Command):
 class cmd_fetch_pack(Command):
     def run(self, argv):
         parser = argparse.ArgumentParser()
-        parser.add_argument('--all', action='store_true')
-        parser.add_argument('location', nargs='?', type=str)
+        parser.add_argument("--all", action="store_true")
+        parser.add_argument("location", nargs="?", type=str)
         args = parser.parse_args(argv)
         client, path = get_transport_and_path(args.location)
         r = Repo(".")
@@ -131,15 +130,24 @@ class cmd_fetch(Command):
         refs = client.fetch(path, r, progress=sys.stdout.write)
         print("Remote refs:")
         for item in refs.items():
-            print("%s -> %s" % item)
+            print("{} -> {}".format(*item))
+
+
+class cmd_for_each_ref(Command):
+    def run(self, args):
+        parser = argparse.ArgumentParser()
+        parser.add_argument("pattern", type=str, nargs="?")
+        args = parser.parse_args(args)
+        for sha, object_type, ref in porcelain.for_each_ref(".", args.pattern):
+            print(f"{sha.decode()} {object_type.decode()}\t{ref.decode()}")
 
 
 class cmd_fsck(Command):
     def run(self, args):
         opts, args = getopt(args, "", [])
         opts = dict(opts)
-        for (obj, msg) in porcelain.fsck("."):
-            print("%s: %s" % (obj, msg))
+        for obj, msg in porcelain.fsck("."):
+            print(f"{obj}: {msg}")
 
 
 class cmd_log(Command):
@@ -174,13 +182,14 @@ class cmd_diff(Command):
 
         r = Repo(".")
         if args == []:
-            commit_id = b'HEAD'
+            commit_id = b"HEAD"
         else:
             commit_id = args[0]
         commit = parse_commit(r, commit_id)
         parent_commit = r[commit.parents[0]]
         porcelain.diff_tree(
-            r, parent_commit.tree, commit.tree, outstream=sys.stdout.buffer)
+            r, parent_commit.tree, commit.tree, outstream=sys.stdout.buffer
+        )
 
 
 class cmd_dump_pack(Command):
@@ -193,18 +202,18 @@ class cmd_dump_pack(Command):
 
         basename, _ = os.path.splitext(args[0])
         x = Pack(basename)
-        print("Object names checksum: %s" % x.name())
-        print("Checksum: %s" % sha_to_hex(x.get_stored_checksum()))
+        print(f"Object names checksum: {x.name()}")
+        print(f"Checksum: {sha_to_hex(x.get_stored_checksum())}")
         if not x.check():
             print("CHECKSUM DOES NOT MATCH")
         print("Length: %d" % len(x))
         for name in x:
             try:
-                print("\t%s" % x[name])
+                print(f"\t{x[name]}")
             except KeyError as k:
-                print("\t%s: Unable to resolve base %s" % (name, k))
+                print(f"\t{name}: Unable to resolve base {k}")
             except ApplyDeltaError as e:
-                print("\t%s: Unable to apply delta: %r" % (name, e))
+                print(f"\t{name}: Unable to apply delta: {e!r}")
 
 
 class cmd_dump_index(Command):
@@ -248,9 +257,28 @@ class cmd_clone(Command):
             "--depth", dest="depth", type=int, help="Depth at which to fetch"
         )
         parser.add_option(
-            "-b", "--branch", dest="branch", type=str,
-            help=("Check out branch instead of branch pointed to by remote "
-                  "HEAD"))
+            "-b",
+            "--branch",
+            dest="branch",
+            type=str,
+            help=("Check out branch instead of branch pointed to by remote " "HEAD"),
+        )
+        parser.add_option(
+            "--refspec",
+            dest="refspec",
+            type=str,
+            help="References to fetch",
+            action="append",
+        )
+        parser.add_option(
+            "--filter",
+            dest="filter_spec",
+            type=str,
+            help="git-rev-list-style object filter",
+        )
+        parser.add_option(
+            "--protocol", dest="protocol", type=int, help="Git protocol version to use"
+        )
         options, args = parser.parse_args(args)
 
         if args == []:
@@ -263,8 +291,19 @@ class cmd_clone(Command):
         else:
             target = None
 
-        porcelain.clone(source, target, bare=options.bare, depth=options.depth,
-                        branch=options.branch)
+        try:
+            porcelain.clone(
+                source,
+                target,
+                bare=options.bare,
+                depth=options.depth,
+                branch=options.branch,
+                refspec=options.refspec,
+                filter_spec=options.filter_spec,
+                protocol_version=options.protocol,
+            )
+        except GitProtocolError as e:
+            print(f"{e}")
 
 
 class cmd_commit(Command):
@@ -300,10 +339,22 @@ class cmd_symbolic_ref(Command):
         porcelain.symbolic_ref(".", ref_name=ref_name, force="--force" in args)
 
 
+class cmd_pack_refs(Command):
+    def run(self, argv):
+        parser = argparse.ArgumentParser()
+        parser.add_argument("--all", action="store_true")
+        # ignored, we never prune
+        parser.add_argument("--no-prune", action="store_true")
+
+        args = parser.parse_args(argv)
+
+        porcelain.pack_refs(".", all=args.all)
+
+
 class cmd_show(Command):
     def run(self, argv):
         parser = argparse.ArgumentParser()
-        parser.add_argument('objectish', type=str, nargs='*')
+        parser.add_argument("objectish", type=str, nargs="*")
         args = parser.parse_args(argv)
         porcelain.show(".", args.objectish or None)
 
@@ -368,7 +419,8 @@ class cmd_reset(Command):
 class cmd_daemon(Command):
     def run(self, args):
         from dulwich import log_utils
-        from dulwich.protocol import TCP_GIT_PORT
+
+        from .protocol import TCP_GIT_PORT
 
         parser = optparse.OptionParser()
         parser.add_option(
@@ -432,7 +484,7 @@ class cmd_write_tree(Command):
     def run(self, args):
         parser = optparse.OptionParser()
         options, args = parser.parse_args(args)
-        sys.stdout.write("%s\n" % porcelain.write_tree("."))
+        sys.stdout.write("{}\n".format(porcelain.write_tree(".")))
 
 
 class cmd_receive_pack(Command):
@@ -471,18 +523,18 @@ class cmd_status(Command):
             for kind, names in status.staged.items():
                 for name in names:
                     sys.stdout.write(
-                        "\t%s: %s\n" % (kind, name.decode(sys.getfilesystemencoding()))
+                        f"\t{kind}: {name.decode(sys.getfilesystemencoding())}\n"
                     )
             sys.stdout.write("\n")
         if status.unstaged:
             sys.stdout.write("Changes not staged for commit:\n\n")
             for name in status.unstaged:
-                sys.stdout.write("\t%s\n" % name.decode(sys.getfilesystemencoding()))
+                sys.stdout.write(f"\t{name.decode(sys.getfilesystemencoding())}\n")
             sys.stdout.write("\n")
         if status.untracked:
             sys.stdout.write("Untracked files:\n\n")
             for name in status.untracked:
-                sys.stdout.write("\t%s\n" % name)
+                sys.stdout.write(f"\t{name}\n")
             sys.stdout.write("\n")
 
 
@@ -494,7 +546,7 @@ class cmd_ls_remote(Command):
             sys.exit(1)
         refs = porcelain.ls_remote(args[0])
         for ref in sorted(refs):
-            sys.stdout.write("%s\t%s\n" % (ref, refs[ref]))
+            sys.stdout.write(f"{ref}\t{refs[ref]}\n")
 
 
 class cmd_ls_tree(Command):
@@ -523,49 +575,64 @@ class cmd_ls_tree(Command):
 
 class cmd_pack_objects(Command):
     def run(self, args):
-        opts, args = getopt(args, "", ["stdout"])
+        deltify = False
+        reuse_deltas = True
+        opts, args = getopt(args, "", ["stdout", "deltify", "no-reuse-deltas"])
         opts = dict(opts)
-        if len(args) < 1 and "--stdout" not in args:
+        if len(args) < 1 and "--stdout" not in opts.keys():
             print("Usage: dulwich pack-objects basename")
             sys.exit(1)
         object_ids = [line.strip() for line in sys.stdin.readlines()]
-        basename = args[0]
-        if "--stdout" in opts:
+        if "--deltify" in opts.keys():
+            deltify = True
+        if "--no-reuse-deltas" in opts.keys():
+            reuse_deltas = False
+        if "--stdout" in opts.keys():
             packf = getattr(sys.stdout, "buffer", sys.stdout)
             idxf = None
             close = []
         else:
-            packf = open(basename + ".pack", "w")
-            idxf = open(basename + ".idx", "w")
+            basename = args[0]
+            packf = open(basename + ".pack", "wb")
+            idxf = open(basename + ".idx", "wb")
             close = [packf, idxf]
-        porcelain.pack_objects(".", object_ids, packf, idxf)
+        porcelain.pack_objects(
+            ".", object_ids, packf, idxf, deltify=deltify, reuse_deltas=reuse_deltas
+        )
         for f in close:
             f.close()
 
 
 class cmd_pull(Command):
     def run(self, args):
-        parser = optparse.OptionParser()
-        options, args = parser.parse_args(args)
-        try:
-            from_location = args[0]
-        except IndexError:
-            from_location = None
-        porcelain.pull(".", from_location)
+        parser = argparse.ArgumentParser()
+        parser.add_argument("from_location", type=str)
+        parser.add_argument("refspec", type=str, nargs="*")
+        parser.add_argument("--filter", type=str, nargs=1)
+        parser.add_argument("--protocol", type=int, nargs=1)
+        args = parser.parse_args(args)
+        porcelain.pull(
+            ".",
+            args.from_location or None,
+            args.refspec or None,
+            filter_spec=args.filter,
+            protocol_version=args.protocol_version or None,
+        )
 
 
 class cmd_push(Command):
-
     def run(self, argv):
         parser = argparse.ArgumentParser()
-        parser.add_argument('-f', '--force', action='store_true', help='Force')
-        parser.add_argument('to_location', type=str)
-        parser.add_argument('refspec', type=str, nargs='*')
+        parser.add_argument("-f", "--force", action="store_true", help="Force")
+        parser.add_argument("to_location", type=str)
+        parser.add_argument("refspec", type=str, nargs="*")
         args = parser.parse_args(argv)
         try:
-            porcelain.push('.', args.to_location, args.refspec or None, force=args.force)
+            porcelain.push(
+                ".", args.to_location, args.refspec or None, force=args.force
+            )
         except porcelain.DivergedBranches:
-            sys.stderr.write('Diverged branches; specify --force to override')
+            sys.stderr.write("Diverged branches; specify --force to override")
             return 1
 
 
@@ -577,26 +644,26 @@ class cmd_remote_add(Command):
 
 
 class SuperCommand(Command):
-
-    subcommands: Dict[str, Type[Command]] = {}
-    default_command: Optional[Type[Command]] = None
+    subcommands: ClassVar[Dict[str, Type[Command]]] = {}
+    default_command: ClassVar[Optional[Type[Command]]] = None
 
     def run(self, args):
         if not args and not self.default_command:
-            print("Supported subcommands: %s" % ", ".join(self.subcommands.keys()))
+            print(
+                "Supported subcommands: {}".format(", ".join(self.subcommands.keys()))
+            )
             return False
         cmd = args[0]
         try:
             cmd_kls = self.subcommands[cmd]
         except KeyError:
-            print("No such subcommand: %s" % args[0])
+            print(f"No such subcommand: {args[0]}")
             return False
         return cmd_kls().run(args[1:])
 
 
 class cmd_remote(SuperCommand):
-
-    subcommands = {
+    subcommands: ClassVar[Dict[str, Type[Command]]] = {
         "add": cmd_remote_add,
     }
 
@@ -606,7 +673,7 @@ class cmd_submodule_list(Command):
         parser = argparse.ArgumentParser()
         parser.parse_args(argv)
         for path, sha in porcelain.submodule_list("."):
-            sys.stdout.write(' %s %s\n' % (sha, path))
+            sys.stdout.write(f" {sha} {path}\n")
 
 
 class cmd_submodule_init(Command):
@@ -617,8 +684,7 @@ class cmd_submodule_init(Command):
 
 
 class cmd_submodule(SuperCommand):
-
-    subcommands = {
+    subcommands: ClassVar[Dict[str, Type[Command]]] = {
         "init": cmd_submodule_init,
     }
 
@@ -670,8 +736,7 @@ class cmd_stash_pop(Command):
 
 
 class cmd_stash(SuperCommand):
-
-    subcommands = {
+    subcommands: ClassVar[Dict[str, Type[Command]]] = {
         "list": cmd_stash_list,
         "pop": cmd_stash_pop,
         "push": cmd_stash_push,
@@ -708,7 +773,7 @@ class cmd_help(Command):
         if options.all:
             print("Available commands:")
             for cmd in sorted(commands):
-                print("  %s" % cmd)
+                print(f"  {cmd}")
         else:
             print(
                 """\
@@ -736,6 +801,7 @@ commands = {
     "dump-index": cmd_dump_index,
     "fetch-pack": cmd_fetch_pack,
     "fetch": cmd_fetch,
+    "for-each-ref": cmd_for_each_ref,
     "fsck": cmd_fsck,
     "help": cmd_help,
     "init": cmd_init,
@@ -744,6 +810,7 @@ commands = {
     "ls-remote": cmd_ls_remote,
     "ls-tree": cmd_ls_tree,
     "pack-objects": cmd_pack_objects,
+    "pack-refs": cmd_pack_refs,
     "pull": cmd_pull,
     "push": cmd_push,
     "receive-pack": cmd_receive_pack,
@@ -770,14 +837,14 @@ def main(argv=None):
         argv = sys.argv[1:]
 
     if len(argv) < 1:
-        print("Usage: dulwich <%s> [OPTIONS...]" % ("|".join(commands.keys())))
+        print("Usage: dulwich <{}> [OPTIONS...]".format("|".join(commands.keys())))
         return 1
 
     cmd = argv[0]
     try:
         cmd_kls = commands[cmd]
     except KeyError:
-        print("No such subcommand: %s" % cmd)
+        print(f"No such subcommand: {cmd}")
         return 1
     # TODO(jelmer): Return non-0 on errors
     return cmd_kls().run(argv[1:])
