@@ -20,16 +20,22 @@
 
 """Object specification."""
 
-from typing import Union, List, Tuple
+from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union
+
+from .objects import Commit, ShaFile, Tree
+
+if TYPE_CHECKING:
+    from .refs import Ref, RefsContainer
+    from .repo import Repo
 
 
-def to_bytes(text):
+def to_bytes(text: Union[str, bytes]) -> bytes:
     if getattr(text, "encode", None) is not None:
-        text = text.encode("ascii")
-    return text
+        text = text.encode("ascii")  # type: ignore
+    return text  # type: ignore
 
 
-def parse_object(repo, objectish):
+def parse_object(repo: "Repo", objectish: Union[bytes, str]) -> "ShaFile":
     """Parse a string referring to an object.
 
     Args:
@@ -43,7 +49,7 @@ def parse_object(repo, objectish):
     return repo[objectish]
 
 
-def parse_tree(repo, treeish):
+def parse_tree(repo: "Repo", treeish: Union[bytes, str]) -> "Tree":
     """Parse a string referring to a tree.
 
     Args:
@@ -64,7 +70,9 @@ def parse_tree(repo, treeish):
     return o
 
 
-def parse_ref(container, refspec):
+def parse_ref(
+    container: Union["Repo", "RefsContainer"], refspec: Union[str, bytes]
+) -> "Ref":
     """Parse a string referring to a reference.
 
     Args:
@@ -89,7 +97,12 @@ def parse_ref(container, refspec):
     raise KeyError(refspec)
 
 
-def parse_reftuple(lh_container, rh_container, refspec, force=False):
+def parse_reftuple(
+    lh_container: Union["Repo", "RefsContainer"],
+    rh_container: Union["Repo", "RefsContainer"],
+    refspec: Union[str, bytes],
+    force: bool = False,
+) -> Tuple[Optional["Ref"], Optional["Ref"], bool]:
     """Parse a reftuple spec.
 
     Args:
@@ -104,6 +117,8 @@ def parse_reftuple(lh_container, rh_container, refspec, force=False):
     if refspec.startswith(b"+"):
         force = True
         refspec = refspec[1:]
+    lh: Optional[bytes]
+    rh: Optional[bytes]
     if b":" in refspec:
         (lh, rh) = refspec.split(b":")
     else:
@@ -125,9 +140,11 @@ def parse_reftuple(lh_container, rh_container, refspec, force=False):
 
 
 def parse_reftuples(
-        lh_container, rh_container,
-        refspecs: Union[bytes, List[bytes], List[Tuple[bytes, bytes]]],
-        force: bool = False):
+    lh_container: Union["Repo", "RefsContainer"],
+    rh_container: Union["Repo", "RefsContainer"],
+    refspecs: Union[bytes, List[bytes]],
+    force: bool = False,
+):
     """Parse a list of reftuple specs to a list of reftuples.
 
     Args:
@@ -167,7 +184,9 @@ def parse_refs(container, refspecs):
     return ret
 
 
-def parse_commit_range(repo, committishs):
+def parse_commit_range(
+    repo: "Repo", committishs: Union[str, bytes]
+) -> Iterator["Commit"]:
     """Parse a string referring to a range of commits.
 
     Args:
@@ -186,19 +205,18 @@ def parse_commit_range(repo, committishs):
 class AmbiguousShortId(Exception):
     """The short id is ambiguous."""
 
-    def __init__(self, prefix, options):
+    def __init__(self, prefix, options) -> None:
         self.prefix = prefix
         self.options = options
 
 
-def scan_for_short_id(object_store, prefix):
+def scan_for_short_id(object_store, prefix, tp):
     """Scan an object store for a short id."""
-    # TODO(jelmer): This could short-circuit looking for objects
-    # starting with a certain prefix.
     ret = []
-    for object_id in object_store:
-        if object_id.startswith(prefix):
-            ret.append(object_store[object_id])
+    for object_id in object_store.iter_prefix(prefix):
+        o = object_store[object_id]
+        if isinstance(o, tp):
+            ret.append(o)
     if not ret:
         raise KeyError(prefix)
     if len(ret) == 1:
@@ -206,7 +224,7 @@ def scan_for_short_id(object_store, prefix):
     raise AmbiguousShortId(prefix, ret)
 
 
-def parse_commit(repo, committish):
+def parse_commit(repo: "Repo", committish: Union[str, bytes]) -> "Commit":
     """Parse a string referring to a single commit.
 
     Args:
@@ -233,7 +251,7 @@ def parse_commit(repo, committish):
             pass
         else:
             try:
-                return scan_for_short_id(repo.object_store, committish)
+                return scan_for_short_id(repo.object_store, committish, Commit)
             except KeyError:
                 pass
     raise KeyError(committish)
