@@ -33,6 +33,7 @@ Currently implemented:
  * describe
  * diff-tree
  * fetch
+ * for-each-ref
  * init
  * ls-files
  * ls-remote
@@ -64,6 +65,7 @@ Functions should generally accept both unicode strings and bytestrings
 """
 
 import datetime
+import fnmatch
 import os
 import posixpath
 import stat
@@ -73,33 +75,62 @@ from collections import namedtuple
 from contextlib import closing, contextmanager
 from io import BytesIO, RawIOBase
 from pathlib import Path
-from typing import Optional, Tuple, Union
+from typing import Dict, List, Optional, Tuple, Union
 
 from .archive import tar_stream
 from .client import get_transport_and_path
 from .config import Config, ConfigFile, StackedConfig, read_submodules
-from .diff_tree import (CHANGE_ADD, CHANGE_COPY, CHANGE_DELETE, CHANGE_MODIFY,
-                        CHANGE_RENAME, RENAME_CHANGE_TYPES)
+from .diff_tree import (
+    CHANGE_ADD,
+    CHANGE_COPY,
+    CHANGE_DELETE,
+    CHANGE_MODIFY,
+    CHANGE_RENAME,
+    RENAME_CHANGE_TYPES,
+)
 from .errors import SendPackError
 from .file import ensure_dir_exists
 from .graph import can_fast_forward
 from .ignore import IgnoreFilterManager
-from .index import (_fs_to_tree_path, blob_from_path_and_stat,
-                    build_file_from_blob, get_unstaged_changes,
-                    index_entry_from_stat)
+from .index import (
+    _fs_to_tree_path,
+    blob_from_path_and_stat,
+    build_file_from_blob,
+    get_unstaged_changes,
+    index_entry_from_stat,
+)
 from .object_store import iter_tree_contents, tree_lookup_path
-from .objects import (Commit, Tag, format_timezone, parse_timezone,
-                      pretty_format_tree_entry)
-from .objectspec import (parse_commit, parse_object, parse_ref,
-                         parse_reftuples, parse_tree, to_bytes)
+from .objects import (
+    Commit,
+    Tag,
+    format_timezone,
+    parse_timezone,
+    pretty_format_tree_entry,
+)
+from .objectspec import (
+    parse_commit,
+    parse_object,
+    parse_ref,
+    parse_reftuples,
+    parse_tree,
+    to_bytes,
+)
 from .pack import write_pack_from_container, write_pack_index
 from .patch import write_tree_diff
 from .protocol import ZERO_SHA, Protocol
-from .refs import (LOCAL_BRANCH_PREFIX, LOCAL_REMOTE_PREFIX, LOCAL_TAG_PREFIX,
-                   _import_remote_refs)
-from .repo import BaseRepo, Repo
-from .server import (FileSystemBackend, ReceivePackHandler, TCPGitServer,
-                     UploadPackHandler)
+from .refs import (
+    LOCAL_BRANCH_PREFIX,
+    LOCAL_REMOTE_PREFIX,
+    LOCAL_TAG_PREFIX,
+    _import_remote_refs,
+)
+from .repo import BaseRepo, Repo, get_user_identity
+from .server import (
+    FileSystemBackend,
+    ReceivePackHandler,
+    TCPGitServer,
+    UploadPackHandler,
+)
 from .server import update_server_info as server_update_server_info
 
 # Module level tuple definition for status output
@@ -130,9 +161,9 @@ DEFAULT_ENCODING = "utf-8"
 
 
 class Error(Exception):
-    """Porcelain-based error. """
+    """Porcelain-based error."""
 
-    def __init__(self, msg):
+    def __init__(self, msg) -> None:
         super().__init__(msg)
 
 
@@ -162,7 +193,7 @@ def parse_timezone_format(tz_str):
     Returns: Timezone offset as integer
     Raises:
       TimezoneFormatError: if timezone information cannot be extracted
-   """
+    """
     import re
 
     # Git internal format
@@ -176,6 +207,7 @@ def parse_timezone_format(tz_str):
 
     # RFC 2822
     import email.utils
+
     rfc_2822 = email.utils.parsedate_tz(tz_str)
     if rfc_2822:
         return rfc_2822[9]
@@ -184,7 +216,9 @@ def parse_timezone_format(tz_str):
 
     # Supported offsets:
     # sHHMM, sHH:MM, sHH
-    iso_8601_pattern = re.compile("[0-9] ?([+-])([0-9]{2})(?::(?=[0-9]{2}))?([0-9]{2})?$")
+    iso_8601_pattern = re.compile(
+        "[0-9] ?([+-])([0-9]{2})(?::(?=[0-9]{2}))?([0-9]{2})?$"
+    )
     match = re.search(iso_8601_pattern, tz_str)
     total_secs = 0
     if match:
@@ -203,7 +237,7 @@ def parse_timezone_format(tz_str):
 def get_user_timezones():
     """Retrieve local timezone as described in
     https://raw.githubusercontent.com/git/git/v2.3.0/Documentation/date-formats.txt
-    Returns: A tuple containing author timezone, committer timezone
+    Returns: A tuple containing author timezone, committer timezone.
     """
     local_timezone = time.localtime().tm_gmtoff
 
@@ -286,7 +320,7 @@ def path_to_tree_path(repopath, path, tree_encoding=DEFAULT_ENCODING):
 class DivergedBranches(Error):
     """Branches have diverged and fast-forward is not possible."""
 
-    def __init__(self, current_sha, new_sha):
+    def __init__(self, current_sha, new_sha) -> None:
         self.current_sha = current_sha
         self.new_sha = new_sha
 
@@ -321,7 +355,6 @@ def archive(
       outstream: Output stream (defaults to stdout)
       errstream: Error stream (defaults to stderr)
     """
-
     if committish is None:
         committish = "HEAD"
     with open_repo_closing(repo) as repo_obj:
@@ -353,7 +386,7 @@ def symbolic_ref(repo, ref_name, force=False):
     with open_repo_closing(repo) as repo_obj:
         ref_path = _make_branch_ref(ref_name)
         if not force and ref_path not in repo_obj.refs.keys():
-            raise Error("fatal: ref `%s` is not a ref" % ref_name)
+            raise Error(f"fatal: ref `{ref_name}` is not a ref")
         repo_obj.refs.set_symbolic_ref(b"HEAD", ref_path)
 
 
@@ -453,6 +486,26 @@ def init(path=".", *, bare=False, symlinks: Optional[bool] = None):
         return Repo.init(path, symlinks=symlinks)
 
 
+def encode_refspecs(refspecs, refspec_encoding):
+    if refspecs is None:
+        return [b"HEAD"]
+
+    def encode_refspec(ref):
+        if isinstance(ref, bytes):
+            return ref
+        else:
+            return ref.encode(refspec_encoding)
+
+    encoded_refs = []
+    if isinstance(refspecs, bytes) or isinstance(refspecs, str):
+        encoded_refs.append(encode_refspec(refspecs))
+    else:
+        for ref in refspecs:
+            encoded_refs.append(encode_refspec(ref))
+
+    return encoded_refs
+
+
 def clone(
     source,
     target=None,
@@ -464,7 +517,11 @@ def clone(
     depth: Optional[int] = None,
     branch: Optional[Union[str, bytes]] = None,
     config: Optional[Config] = None,
-    **kwargs
+    refspecs=None,
+    refspec_encoding=DEFAULT_ENCODING,
+    filter_spec=None,
+    protocol_version: Optional[int] = None,
+    **kwargs,
 ):
     """Clone a local or remote git repository.
 
@@ -480,6 +537,15 @@ def clone(
       branch: Optional branch or tag to be used as HEAD in the new repository
         instead of the cloned repository's HEAD.
       config: Configuration to use
+      refspecs: refspecs to fetch. Can be a bytestring, a string, or a list of
+        bytestring/string.
+      refspec_encoding: Character encoding of bytestrings provided in the refspecs parameter.
+        If not specified, the internal default encoding will be used.
+      filter_spec: A git-rev-list-style object filter spec, as an ASCII string.
+        Only used if the server supports the Git protocol-v2 'filter'
+        feature, and ignored otherwise.
+      protocol_version: desired Git protocol version. By default the highest
+        mutually supported protocol version will be used.
     Returns: The new repository
     """
     if outstream is not None:
@@ -500,6 +566,8 @@ def clone(
     if checkout and bare:
         raise Error("checkout and bare are incompatible")
 
+    encoded_refs = encode_refspecs(refspecs, refspec_encoding)
+
     if target is None:
         target = source.split("/")[-1]
 
@@ -508,8 +576,10 @@ def clone(
 
     mkdir = not os.path.exists(target)
 
-    (client, path) = get_transport_and_path(
-        source, config=config, **kwargs)
+    (client, path) = get_transport_and_path(source, config=config, **kwargs)
+
+    if filter_spec:
+        filter_spec = filter_spec.encode("ascii")
 
     return client.clone(
         path,
@@ -521,6 +591,9 @@ def clone(
         branch=branch,
         progress=errstream.write,
         depth=depth,
+        ref_prefix=encoded_refs,
+        filter_spec=filter_spec,
+        protocol_version=protocol_version,
     )
 
 
@@ -566,7 +639,7 @@ def add(repo=".", paths=None):
 
 
 def _is_subdir(subdir, parentdir):
-    """Check whether subdir is parentdir or a subdir of parentdir
+    """Check whether subdir is parentdir or a subdir of parentdir.
 
     If parentdir or subdir is a relative path, it will be disamgibuated
     relative to the pwd.
@@ -578,7 +651,7 @@ def _is_subdir(subdir, parentdir):
 
 # TODO: option to remove ignored files also, in line with `git clean -fdx`
 def clean(repo=".", target_dir=None):
-    """Remove any untracked files from the target directory recursively
+    """Remove any untracked files from the target directory recursively.
 
     Equivalent to running ``git clean -fd`` in target_dir.
 
@@ -594,9 +667,7 @@ def clean(repo=".", target_dir=None):
             raise Error("target_dir must be in the repo's working dir")
 
         config = r.get_config_stack()
-        require_force = config.get_boolean(  # noqa: F841
-            (b"clean",), b"requireForce", True
-        )
+        config.get_boolean((b"clean",), b"requireForce", True)
 
         # TODO(jelmer): if require_force is set, then make sure that -f, -i or
         # -n is specified.
@@ -640,7 +711,7 @@ def remove(repo=".", paths=None, cached=False):
             try:
                 index_sha = index[tree_path].sha
             except KeyError as exc:
-                raise Error("%s did not match any files" % p) from exc
+                raise Error(f"{p} did not match any files") from exc
 
             if not cached:
                 try:
@@ -663,11 +734,11 @@ def remove(repo=".", paths=None, cached=False):
                         if blob.id != index_sha and index_sha != committed_sha:
                             raise Error(
                                 "file has staged content differing "
-                                "from both the file and head: %s" % p
+                                f"from both the file and head: {p}"
                             )
 
                         if index_sha != committed_sha:
-                            raise Error("file has staged changes: %s" % p)
+                            raise Error(f"file has staged changes: {p}")
                         os.remove(full_path)
             del index[tree_path]
         index.write()
@@ -930,8 +1001,8 @@ def rev_list(repo, commits, outstream=sys.stdout):
 
 
 def _canonical_part(url: str) -> str:
-    name = url.rsplit('/', 1)[-1]
-    if name.endswith('.git'):
+    name = url.rsplit("/", 1)[-1]
+    if name.endswith(".git"):
         name = name[:-4]
     return name
 
@@ -970,10 +1041,10 @@ def submodule_init(repo):
     """
     with open_repo_closing(repo) as r:
         config = r.get_config()
-        gitmodules_path = os.path.join(r.path, '.gitmodules')
+        gitmodules_path = os.path.join(r.path, ".gitmodules")
         for path, url, name in read_submodules(gitmodules_path):
-            config.set((b'submodule', name), b'active', True)
-            config.set((b'submodule', name), b'url', url)
+            config.set((b"submodule", name), b"active", True)
+            config.set((b"submodule", name), b"url", url)
         config.write_to_path()
 
 
@@ -984,6 +1055,7 @@ def submodule_list(repo):
       repo: Path to repository
     """
     from .submodule import iter_cached_submodules
+
     with open_repo_closing(repo) as r:
         for path, sha in iter_cached_submodules(r.object_store, r[r.head()].tree):
             yield path, sha.decode(DEFAULT_ENCODING)
@@ -999,9 +1071,9 @@ def tag_create(
     tag_time=None,
     tag_timezone=None,
     sign=False,
-    encoding=DEFAULT_ENCODING
+    encoding=DEFAULT_ENCODING,
 ):
-    """Creates a tag in git via dulwich calls:
+    """Creates a tag in git via dulwich calls.
 
     Args:
       repo: Path to repository
@@ -1016,7 +1088,6 @@ def tag_create(
         pass True to use default GPG key,
         pass a str containing Key ID to use a specific GPG key)
     """
-
     with open_repo_closing(repo) as r:
         object = parse_object(r, objectish)
 
@@ -1024,8 +1095,7 @@ def tag_create(
             # Create the tag object
             tag_obj = Tag()
             if author is None:
-                # TODO(jelmer): Don't use repo private method.
-                author = r._get_user_identity(r.get_config_stack())
+                author = get_user_identity(r.get_config_stack())
             tag_obj.tagger = author
             tag_obj.message = message + "\n".encode(encoding)
             tag_obj.name = tag
@@ -1074,7 +1144,7 @@ def tag_delete(repo, name):
         elif isinstance(name, list):
             names = name
         else:
-            raise Error("Unexpected tag name type %r" % name)
+            raise Error(f"Unexpected tag name type {name!r}")
         for name in names:
             del r.refs[_make_tag_ref(name)]
 
@@ -1087,7 +1157,6 @@ def reset(repo, mode, treeish="HEAD"):
       mode: Mode ("hard", "soft", "mixed")
       treeish: Treeish to reset to
     """
-
     if mode != "hard":
         raise Error("hard is the only mode currently supported")
 
@@ -1127,9 +1196,9 @@ def push(
     outstream=default_bytes_out_stream,
     errstream=default_bytes_err_stream,
     force=False,
-    **kwargs
+    **kwargs,
 ):
-    """Remote push with dulwich via dulwich.client
+    """Remote push with dulwich via dulwich.client.
 
     Args:
       repo: Path to repository
@@ -1139,7 +1208,6 @@ def push(
       errstream: A stream file to write errors
       force: Force overwriting refs
     """
-
     # Open the repo
     with open_repo_closing(repo) as r:
         if refspecs is None:
@@ -1158,7 +1226,7 @@ def push(
             selected_refs.extend(parse_reftuples(r.refs, refs, refspecs, force=force))
             new_refs = {}
             # TODO: Handle selected_refs == {None: None}
-            for (lh, rh, force_ref) in selected_refs:
+            for lh, rh, force_ref in selected_refs:
                 if lh is None:
                     new_refs[rh] = ZERO_SHA
                     remote_changed_refs[rh] = None
@@ -1166,9 +1234,7 @@ def push(
                     try:
                         localsha = r.refs[lh]
                     except KeyError as exc:
-                        raise Error(
-                            "No valid ref %s in local repository" % lh
-                        ) from exc
+                        raise Error(f"No valid ref {lh} in local repository") from exc
                     if not force_ref and rh in refs:
                         check_diverged(r, refs[rh], localsha)
                     new_refs[rh] = localsha
@@ -1213,28 +1279,38 @@ def pull(
     errstream=default_bytes_err_stream,
     fast_forward=True,
     force=False,
-    **kwargs
+    refspec_encoding=DEFAULT_ENCODING,
+    filter_spec=None,
+    protocol_version=None,
+    **kwargs,
 ):
-    """Pull from remote via dulwich.client
+    """Pull from remote via dulwich.client.
 
     Args:
       repo: Path to repository
       remote_location: Location of the remote
-      refspecs: refspecs to fetch
+      refspecs: refspecs to fetch. Can be a bytestring, a string, or a list of
+        bytestring/string.
       outstream: A stream file to write to output
       errstream: A stream file to write to errors
+      refspec_encoding: Character encoding of bytestrings provided in the refspecs parameter.
+        If not specified, the internal default encoding will be used.
+      filter_spec: A git-rev-list-style object filter spec, as an ASCII string.
+        Only used if the server supports the Git protocol-v2 'filter'
+        feature, and ignored otherwise.
+      protocol_version: desired Git protocol version. By default the highest
+        mutually supported protocol version will be used
     """
     # Open the repo
     with open_repo_closing(repo) as r:
         (remote_name, remote_location) = get_remote_repo(r, remote_location)
 
-        if refspecs is None:
-            refspecs = [b"HEAD"]
+        encoded_refs = encode_refspecs(refspecs, refspec_encoding)
         selected_refs = []
 
         def determine_wants(remote_refs, **kwargs):
             selected_refs.extend(
-                parse_reftuples(remote_refs, r.refs, refspecs, force=force)
+                parse_reftuples(remote_refs, r.refs, encoded_refs, force=force)
             )
             return [
                 remote_refs[lh]
@@ -1245,10 +1321,18 @@ def pull(
         client, path = get_transport_and_path(
             remote_location, config=r.get_config_stack(), **kwargs
         )
+        if filter_spec:
+            filter_spec = filter_spec.encode("ascii")
         fetch_result = client.fetch(
-            path, r, progress=errstream.write, determine_wants=determine_wants
+            path,
+            r,
+            progress=errstream.write,
+            determine_wants=determine_wants,
+            ref_prefix=refspecs,
+            filter_spec=filter_spec,
+            protocol_version=protocol_version,
         )
-        for (lh, rh, force_ref) in selected_refs:
+        for lh, rh, force_ref in selected_refs:
             if not force_ref and rh in r.refs:
                 try:
                     check_diverged(r, r.refs.follow(rh)[1], fetch_result.refs[lh])
@@ -1256,8 +1340,7 @@ def pull(
                     if fast_forward:
                         raise
                     else:
-                        raise NotImplementedError(
-                            "merge is not yet supported") from exc
+                        raise NotImplementedError("merge is not yet supported") from exc
             r.refs[rh] = fetch_result.refs[lh]
         if selected_refs:
             r[b"HEAD"] = fetch_result.refs[selected_refs[0][1]]
@@ -1315,7 +1398,7 @@ def status(repo=".", ignored=False, untracked_files="all"):
 
 
 def _walk_working_dir_paths(frompath, basepath, prune_dirnames=None):
-    """Get path, is_dir for files in working dir from frompath
+    """Get path, is_dir for files in working dir from frompath.
 
     Args:
       frompath: Path to begin walk
@@ -1462,8 +1545,12 @@ def web_daemon(path=".", address=None, port=None):
       address: Optional address to listen on (defaults to ::)
       port: Optional port to listen on (defaults to 80)
     """
-    from .web import (WSGIRequestHandlerLogger, WSGIServerLogger, make_server,
-                      make_wsgi_chain)
+    from .web import (
+        WSGIRequestHandlerLogger,
+        WSGIServerLogger,
+        make_server,
+        make_wsgi_chain,
+    )
 
     backend = FileSystemBackend(path)
     app = make_wsgi_chain(backend)
@@ -1576,7 +1663,7 @@ def branch_create(repo, name, objectish=None, force=False):
             r.refs.set_if_equals(refname, None, object.id, message=ref_message)
         else:
             if not r.refs.add_if_new(refname, object.id, message=ref_message):
-                raise Error("Branch with name %s already exists." % name)
+                raise Error(f"Branch with name {name} already exists.")
 
 
 def branch_list(repo):
@@ -1637,7 +1724,7 @@ def fetch(
     prune=False,
     prune_tags=False,
     force=False,
-    **kwargs
+    **kwargs,
 ):
     """Fetch objects from a remote server.
 
@@ -1673,6 +1760,60 @@ def fetch(
     return fetch_result
 
 
+def for_each_ref(
+    repo: Union[Repo, str] = ".",
+    pattern: Optional[Union[str, bytes]] = None,
+) -> List[Tuple[bytes, bytes, bytes]]:
+    """Iterate over all refs that match the (optional) pattern.
+
+    Args:
+      repo: Path to the repository
+      pattern: Optional glob (7) patterns to filter the refs with
+    Returns:
+      List of bytes tuples with: (sha, object_type, ref_name)
+    """
+    if isinstance(pattern, str):
+        pattern = os.fsencode(pattern)
+
+    with open_repo_closing(repo) as r:
+        refs = r.get_refs()
+
+    if pattern:
+        matching_refs: Dict[bytes, bytes] = {}
+        pattern_parts = pattern.split(b"/")
+        for ref, sha in refs.items():
+            matches = False
+
+            # git for-each-ref uses glob (7) style patterns, but fnmatch
+            # is greedy and also matches slashes, unlike glob.glob.
+            # We have to check parts of the pattern individually.
+            # See https://github.com/python/cpython/issues/72904
+            ref_parts = ref.split(b"/")
+            if len(ref_parts) > len(pattern_parts):
+                continue
+
+            for pat, ref_part in zip(pattern_parts, ref_parts):
+                matches = fnmatch.fnmatchcase(ref_part, pat)
+                if not matches:
+                    break
+
+            if matches:
+                matching_refs[ref] = sha
+
+        refs = matching_refs
+
+    ret: List[Tuple[bytes, bytes, bytes]] = [
+        (sha, r.get_object(sha).type_name, ref)
+        for ref, sha in sorted(
+            refs.items(),
+            key=lambda ref_sha: ref_sha[0],
+        )
+        if ref != b"HEAD"
+    ]
+
+    return ret
+
+
 def ls_remote(remote, config: Optional[Config] = None, **kwargs):
     """List the refs in a remote.
 
@@ -1700,7 +1841,15 @@ def repack(repo):
         r.object_store.pack_loose_objects()
 
 
-def pack_objects(repo, object_ids, packf, idxf, delta_window_size=None, deltify=None, reuse_deltas=True):
+def pack_objects(
+    repo,
+    object_ids,
+    packf,
+    idxf,
+    delta_window_size=None,
+    deltify=None,
+    reuse_deltas=True,
+):
     """Pack objects into a file.
 
     Args:
@@ -1745,7 +1894,7 @@ def ls_tree(
     """
 
     def list_tree(store, treeid, base):
-        for (name, mode, sha) in store[treeid].iteritems():
+        for name, mode, sha in store[treeid].iteritems():
             if base:
                 name = posixpath.join(base, name)
             if name_only:
@@ -1782,7 +1931,7 @@ def remote_add(repo: Repo, name: Union[bytes, str], url: Union[bytes, str]):
 
 
 def remote_remove(repo: Repo, name: Union[bytes, str]):
-    """Remove a remote
+    """Remove a remote.
 
     Args:
       repo: Path to the repository
@@ -1846,8 +1995,7 @@ def update_head(repo, target, detached=False, new_branch=None):
             r.refs.set_symbolic_ref(b"HEAD", to_set)
 
 
-def reset_file(repo, file_path: str, target: bytes = b'HEAD',
-               symlink_fn=None):
+def reset_file(repo, file_path: str, target: bytes = b"HEAD", symlink_fn=None):
     """Reset the file to specific commit or branch.
 
     Args:
@@ -1867,7 +2015,7 @@ def reset_file(repo, file_path: str, target: bytes = b'HEAD',
 
 def _update_head_during_checkout_branch(repo, target):
     checkout_target = None
-    if target == b'HEAD':  # Do not update head while trying to checkout to HEAD.
+    if target == b"HEAD":  # Do not update head while trying to checkout to HEAD.
         pass
     elif target in repo.refs.keys(base=LOCAL_BRANCH_PREFIX):
         update_head(repo, target)
@@ -1879,7 +2027,9 @@ def _update_head_during_checkout_branch(repo, target):
         if config.has_section(section):
             checkout_target = target.replace(name + b"/", b"")
             try:
-                branch_create(repo, checkout_target, (LOCAL_REMOTE_PREFIX + target).decode())
+                branch_create(
+                    repo, checkout_target, (LOCAL_REMOTE_PREFIX + target).decode()
+                )
             except Error:
                 pass
             update_head(repo, LOCAL_BRANCH_PREFIX + checkout_target)
@@ -1890,7 +2040,7 @@ def _update_head_during_checkout_branch(repo, target):
 
 
 def checkout_branch(repo, target: Union[bytes, str], force: bool = False):
-    """switch branches or restore working tree files.
+    """Switch branches or restore working tree files.
 
     The implementation of this function will probably not scale well
     for branches with lots of local changes.
@@ -1912,7 +2062,14 @@ def checkout_branch(repo, target: Union[bytes, str], force: bool = False):
         _update_head_during_checkout_branch(repo, target)
     else:
         status_report = status(repo)
-        changes = list(set(status_report[0]['add'] + status_report[0]['delete'] + status_report[0]['modify'] + status_report[1]))
+        changes = list(
+            set(
+                status_report[0]["add"]
+                + status_report[0]["delete"]
+                + status_report[0]["modify"]
+                + status_report[1]
+            )
+        )
         index = 0
         while index < len(changes):
             change = changes[index]
@@ -1922,7 +2079,10 @@ def checkout_branch(repo, target: Union[bytes, str], force: bool = False):
                     target_tree.lookup_path(repo.object_store.__getitem__, change)
                     index += 1
                 except KeyError:
-                    raise CheckoutError('Your local changes to the following files would be overwritten by checkout: ' + change.decode())
+                    raise CheckoutError(
+                        "Your local changes to the following files would be overwritten by checkout: "
+                        + change.decode()
+                    )
             except KeyError:
                 changes.pop(index)
 
@@ -1941,7 +2101,7 @@ def checkout_branch(repo, target: Union[bytes, str], force: bool = False):
             blob = repo.object_store[entry.sha]
             ensure_dir_exists(os.path.dirname(full_path))
             st = build_file_from_blob(blob, entry.mode, full_path)
-            repo_index[entry.path] = index_entry_from_stat(st, entry.sha, 0)
+            repo_index[entry.path] = index_entry_from_stat(st, entry.sha)
 
         repo_index.write()
 
